//
//  IRController.m
//  IR
//
//  Created by Phidgets on 2015-07-28.
//  Copyright (c) 2015 Phidgets. All rights reserved.
//

#import "IRController.h"

#define RAWDATA_LONGSPACE  -1

@implementation IRController

- (void)windowWillClose:(NSNotification*)aNotification{
	if(ch){
		//Safely close example
		Phidget_setOnAttachHandler((PhidgetHandle)ch, NULL, NULL);
		Phidget_setOnDetachHandler((PhidgetHandle)ch, NULL, NULL);
		Phidget_setOnErrorHandler((PhidgetHandle)ch, NULL, NULL);
		Phidget_close((PhidgetHandle)ch);
		PhidgetIR_delete(&ch);
		ch = NULL;
	}
}


-(void)awakeFromNib{
    PhidgetReturnCode result;
    
    //Manage GUI
    NSRect frame = [mainWindow frame];
    frame.size.height = [mainWindow minSize].height;
    [mainWindow setFrame:frame display:NO];
    [mainWindow center];
    [mainWindow setDelegate:self];
    [phidgetInfoBoxView fillPhidgetInfo:nil];
    [phidgetInfoBox setHidden:NO];
    
    result = PhidgetIR_create(&ch);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    
    result = [self initChannel:(PhidgetHandle)ch];
    if(result != EPHIDGET_OK){
        const char* resultString;
        Phidget_getErrorDescription(result, &resultString);
        NSString *errorString = [NSString stringWithFormat:@"Failed to initialize channel: %@",  [NSString stringWithUTF8String:resultString]];
        [self outputError:[errorString UTF8String]];
    }
    
    result = PhidgetIR_setOnCodeHandler(ch, gotCodeUpdate, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    
    result = PhidgetIR_setOnLearnHandler(ch, gotLearnUpdate, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    
    result = PhidgetIR_setOnRawDataHandler(ch, gotRawData, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self outputLastError];
    }
    
	[PhidgetInfoBox openCmdLine:(PhidgetHandle)ch];
}

-(PhidgetReturnCode)initChannel:(PhidgetHandle) channel{
    PhidgetReturnCode result;
    
    result = Phidget_setOnAttachHandler(channel, gotAttach, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return result;
    }
    
    result = Phidget_setOnDetachHandler(channel, gotDetach, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return result;
    }
    
    result = Phidget_setOnErrorHandler(channel, gotError, (__bridge void*)self);
    if(result != EPHIDGET_OK){
        [self outputLastError];
        return result;
    }
    
    /*
     * Please review the Phidget22 channel matching documentation for details on the device
     * and class architecture of Phidget22, and how channels are matched to device features.
     */
    
    /*
     * Specifies the serial number of the device to attach to.
     * For VINT devices, this is the hub serial number.
     *
     * The default is any device.
     */
    // Phidget_setDeviceSerialNumber(ch, <YOUR DEVICE SERIAL NUMBER>);
    
    /*
     * For VINT devices, this specifies the port the VINT device must be plugged into.
     *
     * The default is any port.
     */
    // Phidget_setHubPort(ch, 0);
    
    /*
     * Specifies that the channel should only match a VINT hub port.
     * The only valid channel id is 0.
     *
     * The default is 0 (false), meaning VINT hub ports will never match
     */
    // Phidget_setIsHubPortDevice(ch, 1);
    
    /*
     * Specifies which channel to attach to.  It is important that the channel of
     * the device is the same class as the channel that is being opened.
     *
     * The default is any channel.
     */
    // Phidget_setChannel(ch, 0);
    
    /*
     * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
     * In a normal envchonment this can be done automatically by enabling server discovery, which
     * will cause the client to discovery and connect to available servers.
     *
     * To force the channel to only match a network Phidget, set remote to 1.
     */
    // PhidgetNet_enableServerDiscovery(PHIDGETSERVER_DEVICEREMOTE);
    // Phidget_setIsRemote(ch, 1);
    
    return EPHIDGET_OK;
}

#pragma mark Event callbacks
static void gotAttach(PhidgetHandle phid, void *context) {
	[(__bridge id)context performSelectorOnMainThread:@selector(onAttachHandler)
										   withObject:nil
										waitUntilDone:NO];
}
static void gotDetach(PhidgetHandle phid, void *context) {
	[(__bridge id)context performSelectorOnMainThread:@selector(onDetachHandler)
										   withObject:nil
										waitUntilDone:NO];
}
static void gotError(PhidgetHandle phid, void *context, Phidget_ErrorEventCode errcode, const char *error) {
	[(__bridge id)context performSelectorOnMainThread:@selector(errorHandler:)
										   withObject:[NSArray arrayWithObjects:[NSNumber numberWithInt:errcode], [NSString stringWithUTF8String:error], nil]
										waitUntilDone:NO];
}
void gotCodeUpdate(PhidgetIRHandle phid, void *context, const char *codeStr, uint32_t bitCount, int repeat) {
	@autoreleasepool {
		NSMutableArray *dataArray = [[NSMutableArray alloc] initWithCapacity:3];
		[dataArray insertObject:[NSString stringWithCString:codeStr encoding:NSASCIIStringEncoding] atIndex:0];
		[dataArray insertObject:[NSNumber numberWithInt:bitCount] atIndex:1];
		[dataArray insertObject:[NSNumber numberWithInt:repeat] atIndex:2];
		
		[(__bridge id)context performSelectorOnMainThread:@selector(onCodeUpdateHandler:)
											   withObject:dataArray
											waitUntilDone:NO];
	}
	
}
void gotLearnUpdate(PhidgetIRHandle phid, void *context, const char *codeStr, PhidgetIR_CodeInfoHandle codeInfo) {
	@autoreleasepool {
		NSMutableArray *dataArray = [[NSMutableArray alloc] initWithCapacity:2];
		[dataArray insertObject:[NSString stringWithCString:codeStr encoding:NSASCIIStringEncoding] atIndex:0];
		[dataArray insertObject:[[PhidgetIRCodeInfo alloc] initWithHandle:codeInfo] atIndex:1];
		
		[(__bridge id)context performSelectorOnMainThread:@selector(onLearnUpdateHandler:)
											   withObject:dataArray
											waitUntilDone:NO];
	}
	
}
void gotRawData(PhidgetIRHandle phid, void *context, const uint32_t *data, size_t dataLength) {
	@autoreleasepool {
		NSMutableArray *dataArray = [[NSMutableArray alloc] initWithCapacity:dataLength + 1];
		[dataArray insertObject:[NSNumber numberWithInt:(int)dataLength] atIndex:0];
		for(int i = 0; i < dataLength; i ++)
			[dataArray addObject:[NSNumber numberWithInt:data[i]]];
		
		[(__bridge id)context performSelectorOnMainThread:@selector(onRawDataHandler:)
											   withObject:dataArray
											waitUntilDone:NO];
	}
}

#pragma mark Attach, detach, data, and error events
- (void)onAttachHandler{
	//Adjust GUI based on information from channel
	[phidgetInfoBoxView fillPhidgetInfo:(PhidgetHandle)ch];
	[codeBox setHidden:NO];
	[learnBox setHidden:NO];
	[showRawDataCheck setIntValue:0];
	[retransmitButton setEnabled:NO];
	[transmitRawDataButton setEnabled:NO];
    
    //Adjusting window height
    NSRect frame = [mainWindow frame];
    frame.size.height = [mainWindow maxSize].height;
    [mainWindow setFrame:frame display:NO];
    [mainWindow center];
}

- (void)onDetachHandler{
	//Reset
	[codeBox setHidden:YES];
	[learnBox setHidden:YES];
	[phidgetInfoBoxView fillPhidgetInfo:nil];
    
    //Adjusting window height
    NSRect frame = [mainWindow frame];
    frame.size.height = [mainWindow minSize].height;
    [mainWindow setFrame:frame display:NO];
    [mainWindow center];
}


static int errorCounter = 0;
-(void) outputError:(const char *)errorString{
    errorCounter++;
    NSAttributedString *outputString = [[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"%@\n",[NSString stringWithUTF8String:errorString]]
    	attributes:@{NSForegroundColorAttributeName: NSColor.controlTextColor}];
    [[errorEventLog textStorage] beginEditing];
    [[errorEventLog textStorage] appendAttributedString:outputString];
    [[errorEventLog textStorage] endEditing];
    
    [errorEventLogCounter setIntValue:errorCounter];
    if(![errorEventLogWindow isVisible])
        [errorEventLogWindow setIsVisible:YES];
}

-(void)errorHandler:(NSArray *)errorEventData{
    const char* errorString = [[errorEventData objectAtIndex:1] UTF8String];
    [self outputError:errorString];
}

-(void)outputLastError {
	const char *errorString;
	char *errorDetailString = NULL;
    size_t errorDetailStringLen = 0;
    PhidgetReturnCode lastErr;
	
    if (!Phidget_getLastError(&lastErr, &errorString, NULL, &errorDetailStringLen)) {
		errorDetailString = malloc(errorDetailStringLen);
		Phidget_getLastError(&lastErr, &errorString, errorDetailString, &errorDetailStringLen);
    	[self outputError:errorDetailString];
    	free(errorDetailString);
    }
}

int repeatCount = 0;
- (void)onCodeUpdateHandler:(NSArray *)dataArray{
	NSString *irCode = @"0x";
	int bitCount = [[dataArray objectAtIndex:1] intValue];
	int repeat = [[dataArray objectAtIndex:2] intValue];
	
	repeatCount++;
	if(repeat == 0)
		repeatCount = 0;
    
    irCode = [irCode stringByAppendingString:[dataArray objectAtIndex:0]];
    
	[bitLength0Text setStringValue:[NSString stringWithFormat:@"%d",bitCount]];
	[repeatCodeText setStringValue:[NSString stringWithFormat:@"%d",repeatCount]];
	[irCode0Text setStringValue:irCode];
}

- (void)onLearnUpdateHandler:(NSArray *)dataArray{
	NSString *irCode = @"0x";
	NSString *toggleMaskString = @"0x";
	NSString *repeatString = @"";
	
	irCode = [irCode stringByAppendingString:[dataArray objectAtIndex:0]];
	PhidgetIR_CodeInfo *codeInfo = [[dataArray objectAtIndex:1]CodeInfo];
	int byteLength = codeInfo->bitCount/8;
	byteLength = codeInfo->bitCount%8 ? byteLength++ : byteLength;
	
	for(int i = 0; i < 26; i ++){
		if(codeInfo->repeat[i])
			repeatString = [repeatString stringByAppendingString:[NSString stringWithFormat:@"%d,",codeInfo->repeat[i]]];
		else
			break;
	}
	
	for(int i = 1; i <= byteLength; i++)
		toggleMaskString = [toggleMaskString stringByAppendingString:[NSString stringWithFormat:@"%02x",codeInfo->toggleMask[i-1]]];
	
	switch(codeInfo->encoding){
		case IR_ENCODING_SPACE:
			[encodingText setStringValue:@"Space"];
			break;
		case IR_ENCODING_PULSE:
			[encodingText setStringValue:@"Pulse"];
			break;
		case IR_ENCODING_BIPHASE:
			[encodingText setStringValue:@"BiPhase"];
			break;
		case IR_ENCODING_RC5:
			[encodingText setStringValue:@"RC5"];
			break;
		case IR_ENCODING_RC6:
			[encodingText setStringValue:@"RC6"];
			break;
		default:
			[encodingText setStringValue:@"Error"];
			break;
	}
	
	[irCode1Text setStringValue:irCode];
    [bitLength1Text setStringValue:[NSString stringWithFormat:@"%@",codeInfo->length == 1 ? @"Unknown" : codeInfo->length == 2 ? @"Constant" : @"Variable"]];
	[gapText setStringValue:[NSString stringWithFormat:@"%d",codeInfo->gap]];
	[trailText setStringValue:[NSString stringWithFormat:@"%d",codeInfo->trail]];
	[zeroText setStringValue:[NSString stringWithFormat:@"%d,%d",codeInfo->zero[0],codeInfo->zero[1]]];
	[oneText setStringValue:[NSString stringWithFormat:@"%d,%d",codeInfo->one[0],codeInfo->one[1]]];
	[headerText setStringValue:[NSString stringWithFormat:@"%d,%d",codeInfo->header[0],codeInfo->header[1]]];
	[repeatText setStringValue:repeatString];
	[toggleMaskText setStringValue:toggleMaskString];
	[retransmitButton setEnabled:YES];
}
int rawDataCol = 0;
int rawDataBuffer[199];
int rawDataCount = 0;
-(void)onRawDataHandler:(NSArray *) dataArray{
	int length = [[dataArray objectAtIndex:0] intValue];
	if([showRawDataCheck intValue] == 1){
		NSAttributedString *logString;
  
		for(int i = 1; i < length+1; i++){
			if([[dataArray objectAtIndex:i]intValue] == RAWDATA_LONGSPACE){
				logString = [[NSAttributedString alloc] initWithString:@"\n\nLONG\n"
    				attributes:@{NSForegroundColorAttributeName: NSColor.controlTextColor}];
				rawDataCol = -1;
				rawDataCount = 0;
			}
			else if([[dataArray objectAtIndex:i]intValue] > 100000){
				logString = [[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"\n\n%10d\n",[[dataArray objectAtIndex:i]intValue]]
    				attributes:@{NSForegroundColorAttributeName: NSColor.controlTextColor}];
				rawDataCol = -1;
				rawDataBuffer[rawDataCount++] = [[dataArray objectAtIndex:i]intValue];
			}
			else{
				
				if(rawDataCol == 9){
					logString = [[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"%10d\n",[[dataArray objectAtIndex:i]intValue]]
    					attributes:@{NSForegroundColorAttributeName: NSColor.controlTextColor}];
					rawDataCol = -1;
				}
				else
					logString = [[NSAttributedString alloc] initWithString:[NSString stringWithFormat:@"%10d,",[[dataArray objectAtIndex:i]intValue]]
    					attributes:@{NSForegroundColorAttributeName: NSColor.controlTextColor}];
                
				rawDataBuffer[rawDataCount++] = [[dataArray objectAtIndex:i]intValue];
			}
			rawDataCol++;
			[[rawDataLog textStorage] appendAttributedString:logString];
			[rawDataLog scrollRangeToVisible:NSMakeRange([[rawDataLog string] length], 0)];
			if(rawDataCount > 0)
				[transmitRawDataButton setEnabled:YES];
			else
				[transmitRawDataButton setEnabled:YES];
            
            if(rawDataCount > 199)
                rawDataCount = 0;
		}
	}
}
#pragma mark GUI controls
- (IBAction)retransmit:(id)sender{
	PhidgetIR_CodeInfo codeInfo;
	char code[16];
	size_t codeLength = 16;
	
	PhidgetIR_getLastLearnedCode(ch, code, codeLength, &codeInfo);
	PhidgetIR_transmit(ch, code, &codeInfo);
}
- (IBAction)retransmitRaw:(id)sender{
    [self performSelectorInBackground:@selector(transmitBackground) withObject:nil];
}
-(void)transmitBackground{
    uint32 sendArray[199];
    int j = 0;
    for(int i = 0; i < rawDataCount; i++){
        if(rawDataBuffer[i] >= 50000){
            PhidgetIR_transmitRaw(ch, sendArray, j, 0, 0, rawDataBuffer[i]);
            j = 0;
        }
        else{
            sendArray[j++] = rawDataBuffer[i];
        }
    }
    PhidgetIR_transmitRaw(ch, sendArray, j, 0, 0, 150000);
}
- (IBAction)showRAWData:(id)sender{
	[rawDataLog setString:@""];
	if([showRawDataCheck intValue]){
		//bring items into view
		[rawDataBox setHidden:NO];
		//Adjusting view height
		NSRect frame = [mainWindow frame];
		frame.size.width += 675;
		[mainWindow setFrame:frame display:YES animate:YES];
	}
	else{
		//Adjusting view height
		NSRect frame = [mainWindow frame];
		frame.size.width -= 675;
		[mainWindow setFrame:frame display:YES animate:YES];
		//take items out of view
		[rawDataBox setHidden:YES];
	}
}
- (IBAction)clear:(id)sender{
	[rawDataLog setString:@""];
	rawDataCount = 0;
	[transmitRawDataButton setEnabled:NO];
}

@end
